# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding=utf-8

from webullsdkmdata.quotes.subscribe.basic_quote_decoder import BasicQuoteDecoder
from webullsdkmdata.quotes.subscribe.payload_type import (PAYLOAD_TYPE_BASIC_QUOTE,
                                                PAYLOAD_TYPE_SHAPSHOT)
from webullsdkmdata.quotes.subscribe.snapshot_decoder import SnapshotDecoder
from webullsdkmdata.request.get_streaming_token_request import \
    GetStreamingTokenRequest
from webullsdkmdata.request.quotes_subscribe_request import SubscribeRequest
from webullsdkquotescore.quotes_client import QuotesClient
from webullsdkquotescore.quotes_client import LOG_ERR


class DefaultQuotesClient(QuotesClient):
    def __init__(self, app_key, app_secret, region_id, api_endpoint=None, tls_enable=True, transport="tcp", retry_policy=None):
        if region_id:
            super().__init__(app_key, app_secret, region_id,
                             tls_enable, transport, retry_policy)
        else:
            super().__init__(app_key, app_secret, tls_enable=tls_enable,
                             transport=transport, retry_policy=retry_policy)
        if api_endpoint:
            self._api_endpoint = api_endpoint
        self._on_subscribe_success = None

    @property
    def on_subscribe_success(self):
        return self._on_subscribe_success

    @on_subscribe_success.setter
    def on_subscribe_success(self, func):
        with self._callback_mutex:
            self._on_subscribe_success = func

    def _default_refresh_token_func(self, client, api_client):
        request = GetStreamingTokenRequest()
        if client._api_endpoint:
            request.set_endpoint(client._api_endpoint)
        response = api_client.get_response(request)
        return response.json()['token']

    def _default_quotes_subscribe_func(self, client, api_client, token):
        request = SubscribeRequest()
        if client._api_endpoint:
           request.set_endpoint(client._api_endpoint)
        request.set_token(token)
        request.set_symbols(self.quotes_symbols)
        request.set_category(self.quotes_category)
        request.set_subscribe_types(self.quotes_subtypes)
        response = api_client.get_response(request)
        if response.status_code == 200:
            with self._callback_mutex:
                _on_subscribe_success = self._on_subscribe_success
                if _on_subscribe_success:
                    try:
                        _on_subscribe_success(client, api_client, token)
                    except Exception as e:
                        self._easy_log(
                            LOG_ERR, 'Caught exception in on_subscribe_success: %s', e)
                        raise

    def init_default_settings(self, symbols, category, sub_types):
        self.on_refresh_token = self._default_refresh_token_func
        self.on_quotes_subscribe = self._default_quotes_subscribe_func
        self.register_payload_decoder(
            PAYLOAD_TYPE_BASIC_QUOTE, BasicQuoteDecoder())
        self.register_payload_decoder(PAYLOAD_TYPE_SHAPSHOT, SnapshotDecoder())
        self.quotes_symbols = symbols
        self.quotes_category = category
        self.quotes_subtypes = sub_types



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: message.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='message.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\rmessage.proto\"E\n\tBasicData\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x15\n\rinstrument_id\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\t\"p\n\tQuoteData\x12\x19\n\x05\x62\x61sic\x18\x01 \x01(\x0b\x32\n.BasicData\x12\x10\n\x08\x61sk_size\x18\x02 \x01(\t\x12\x11\n\task_price\x18\x03 \x01(\t\x12\x10\n\x08\x62id_size\x18\x04 \x01(\t\x12\x11\n\tbid_price\x18\x05 \x01(\t\"\xf4\x01\n\x0cSnapshotData\x12\x19\n\x05\x62\x61sic\x18\x01 \x01(\x0b\x32\n.BasicData\x12\x0c\n\x04open\x18\x02 \x01(\t\x12\x0c\n\x04high\x18\x03 \x01(\t\x12\x0b\n\x03low\x18\x04 \x01(\t\x12\r\n\x05price\x18\x05 \x01(\t\x12\x11\n\tpre_close\x18\x06 \x01(\t\x12\x0e\n\x06volume\x18\x07 \x01(\t\x12\x0e\n\x06\x63hange\x18\x08 \x01(\t\x12\x14\n\x0c\x63hange_ratio\x18\t \x01(\t\x12\x10\n\x08\x61sk_size\x18\n \x01(\t\x12\x11\n\task_price\x18\x0b \x01(\t\x12\x10\n\x08\x62id_size\x18\x0c \x01(\t\x12\x11\n\tbid_price\x18\r \x01(\tb\x06proto3'
)




_BASICDATA = _descriptor.Descriptor(
  name='BasicData',
  full_name='BasicData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='symbol', full_name='BasicData.symbol', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='instrument_id', full_name='BasicData.instrument_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='BasicData.timestamp', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17,
  serialized_end=86,
)


_QUOTEDATA = _descriptor.Descriptor(
  name='QuoteData',
  full_name='QuoteData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='basic', full_name='QuoteData.basic', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ask_size', full_name='QuoteData.ask_size', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ask_price', full_name='QuoteData.ask_price', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bid_size', full_name='QuoteData.bid_size', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bid_price', full_name='QuoteData.bid_price', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=88,
  serialized_end=200,
)


_SNAPSHOTDATA = _descriptor.Descriptor(
  name='SnapshotData',
  full_name='SnapshotData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='basic', full_name='SnapshotData.basic', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='open', full_name='SnapshotData.open', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='high', full_name='SnapshotData.high', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='low', full_name='SnapshotData.low', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='price', full_name='SnapshotData.price', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pre_close', full_name='SnapshotData.pre_close', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='volume', full_name='SnapshotData.volume', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='change', full_name='SnapshotData.change', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='change_ratio', full_name='SnapshotData.change_ratio', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ask_size', full_name='SnapshotData.ask_size', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ask_price', full_name='SnapshotData.ask_price', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bid_size', full_name='SnapshotData.bid_size', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bid_price', full_name='SnapshotData.bid_price', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=203,
  serialized_end=447,
)

_QUOTEDATA.fields_by_name['basic'].message_type = _BASICDATA
_SNAPSHOTDATA.fields_by_name['basic'].message_type = _BASICDATA
DESCRIPTOR.message_types_by_name['BasicData'] = _BASICDATA
DESCRIPTOR.message_types_by_name['QuoteData'] = _QUOTEDATA
DESCRIPTOR.message_types_by_name['SnapshotData'] = _SNAPSHOTDATA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

BasicData = _reflection.GeneratedProtocolMessageType('BasicData', (_message.Message,), {
  'DESCRIPTOR' : _BASICDATA,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:BasicData)
  })
_sym_db.RegisterMessage(BasicData)

QuoteData = _reflection.GeneratedProtocolMessageType('QuoteData', (_message.Message,), {
  'DESCRIPTOR' : _QUOTEDATA,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:QuoteData)
  })
_sym_db.RegisterMessage(QuoteData)

SnapshotData = _reflection.GeneratedProtocolMessageType('SnapshotData', (_message.Message,), {
  'DESCRIPTOR' : _SNAPSHOTDATA,
  '__module__' : 'message_pb2'
  # @@protoc_insertion_point(class_scope:SnapshotData)
  })
_sym_db.RegisterMessage(SnapshotData)


# @@protoc_insertion_point(module_scope)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding=utf-8

from webullsdkcore.request import ApiRequest

class SubscribeRequest(ApiRequest):
    def __init__(self):
        ApiRequest.__init__(self, "/market-data/streaming/subscribe", version='v1', body_params={})
        
    def set_token(self, token):
        self.add_body_params("token", token)

    def set_symbols(self, symbols):
        if isinstance(symbols, str):
            self.add_body_params("symbols", symbols.split(","))
            return
        elif isinstance(symbols, list):
            self.add_body_params("symbols", symbols)
            return
        raise ValueError("not supported type")

    def set_category(self, category):
        self.add_body_params("category", category)
        
    def set_subscribe_types(self, types):
        if isinstance(types, str):
            self.add_body_params("sub_types", types.split(","))
            return
        elif isinstance(types, list):
            self.add_body_params("sub_types", types)
            return
        raise ValueError("not supported type")





# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from webullsdkmdata.request.get_historical_bars_request import GetHistoricalBarsRequest
from webullsdkmdata.request.get_quote_request import GetQuoteRequest
from webullsdkmdata.request.get_streaming_token_request import GetStreamingTokenRequest
from webullsdkmdata.request.quotes_subscribe_request import SubscribeRequest
from webullsdkmdata.request.quotes_unsubscribe_request import UnsubcribeRequest



class MarketData:
    def __init__(self, api_client):
        self.client = api_client

    def get_token(self):
        token_request = GetStreamingTokenRequest()
        response = self.client.get_response(token_request)
        return response

    def create_subscription_rel(self, token, symbols, category, sub_types):
        subscribe_request = SubscribeRequest()
        subscribe_request.set_token(token)
        subscribe_request.set_symbols(symbols)
        subscribe_request.set_category(category)
        subscribe_request.set_subscribe_types(sub_types)
        response = self.client.get_response(subscribe_request)
        return response

    def remove_subscription_rel(self, token, symbols, category, sub_types, unsubscribe_all=False):
        unsubscribe_request = UnsubcribeRequest()
        unsubscribe_request.set_token(token)
        if unsubscribe_all:
            unsubscribe_request.set_unsubscribe_all(is_all=unsubscribe_all)
        else:
            unsubscribe_request.set_symbols(symbols)
            unsubscribe_request.set_category(category)
            unsubscribe_request.set_sub_types(sub_types)
        response = self.client.get_response(unsubscribe_request)
        return response

    def get_history_bar(self, symbol, category, timespan, count='200'):
        history_bar_request = GetHistoricalBarsRequest()
        history_bar_request.set_symbol(symbol)
        history_bar_request.set_category(category)
        history_bar_request.set_timespan(timespan)
        history_bar_request.set_count(count)
        response = self.client.get_response(history_bar_request)
        return response

    def get_quote(self, symbols, category):
        quote_request = GetQuoteRequest()
        quote_request.set_symbols(symbols)
        quote_request.set_category(category)
        response = self.client.get_response(quote_request)
        return response



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding=utf-8

from decimal import Decimal
from webullsdkmdata.quotes.subscribe.basic_result import BasicResult


class BasicQuoteResult:
    def __init__(self, pb_quote):
        self.basic = BasicResult(pb_quote.basic)
        self.ask_size = Decimal(
            pb_quote.ask_size) if pb_quote.ask_size else None
        self.ask_price = Decimal(
            pb_quote.ask_price) if pb_quote.ask_price else None
        self.bid_size = Decimal(
            pb_quote.bid_size) if pb_quote.bid_size else None
        self.bid_price = Decimal(
            pb_quote.bid_price) if pb_quote.bid_price else None

    def get_basic(self):
        return self.basic

    def get_ask_size(self):
        return self.ask_size

    def get_ask_price(self):
        return self.ask_price

    def get_bid_size(self):
        return self.bid_size

    def get_bid_price(self):
        return self.bid_price

    def __repr__(self):
        return "%s, ask_size:%s, ask_price:%s, bid_size:%s, bid_price:%s" \
            % (self.basic, self.ask_size, self.ask_price, self.bid_size, self.bid_price)

    def __str__(self):
        return self.__repr__()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding=utf-8
from datetime import datetime


class BasicResult:
    def __init__(self, pb_basic):
        self.symbol = pb_basic.symbol
        self.instrument_id = pb_basic.instrument_id
        self.timestamp = int(pb_basic.timestamp)

    def get_symbol(self):
        return self.symbol

    def get_instrument_id(self):
        return self.instrument_id

    def get_timestmap(self):
        return self.timestamp

    def get_timestamp_as_utc(self):
        return datetime.utcfromtimestamp(self.timestamp/1000.0)

    def __repr__(self):
        return "%s-%s-%d" % (self.symbol, self.instrument_id, self.timestamp)

    def __str__(self):
        return self.__repr__()
